// FirstUXRes.cpp : main project file.

#include "stdafx.h"

using namespace System;
using namespace System::IO;
using namespace System::Drawing;

// Simple structure for holding color values as 4 doubles
struct RGBA { double R, G, B, A; };

/**
This algorithm for converting a color to alpha is based on the one used by GIMP
http://git.gnome.org/browse/gimp/tree/plug-ins/common/color-to-alpha.c
**/
double color2alpha_single(const double src, const double color) {
	if (color < 0.0001)		return src;
	else if (src > color)	return (src - color) / (1.0 - color);
	else if (src < color)	return (color - src) / color;
	else					return 0.0;
}
RGBA color2alpha(RGBA src, const RGBA color) { // src is color from image, color is color to replace
	RGBA x = {
		color2alpha_single(src.R, color.R),
		color2alpha_single(src.G, color.G),
		color2alpha_single(src.B, color.B),
		src.A
	};
	if (x.R > x.G)	src.A = (x.R > x.B) ? x.R : x.B;
	else			src.A = (x.G > x.B) ? x.G : x.B;
	if (src.A >= 0.0001) {
		src.R = (src.R - color.R) / src.A + color.R;
		src.G = (src.G - color.G) / src.A + color.G;
		src.B = (src.B - color.B) / src.A + color.B;
		src.A *= x.A;
	}
	return src;
}

// Mixes two colors based on their alpha value, placing a over b.
// Note: does not mix alpha values, simply sets the alpha to 1.0
RGBA mixColors(RGBA a, const RGBA b) {
	a.R = a.R*a.A + b.R*b.A*(1.0-a.A);
	a.G = a.G*a.A + b.G*b.A*(1.0-a.A);
	a.B = a.B*a.A + b.B*b.A*(1.0-a.A);
	a.A = 1.0;
	return a;
}

// Converts RGBA to System::Drawing::Color
Color rgba2color(RGBA x) {
	return Color::FromArgb((int)(x.A*255), (int)(x.R*255), (int)(x.G*255), (int)(x.B*255));
}

// Converts System::Drawing::Color to RGBA
RGBA color2rgba(Color c) {
	RGBA x = { c.R/255.0, c.G/255.0, c.B/255.0, c.A/255.0 };
	return x;
}

int main(array<System::String ^> ^args) {
	if (args->Length != 3) {
		Console::WriteLine(L"You need to include the background image (PNG, GIF, JPEG, EXIG, or TIFF) file, a source directory (containing all the FirstUXRes animation files), and a destination directory on the command line.");
		return 1;
	}

	// Color to replace
	RGBA black = {0.0, 0.0, 0.0, 1.0};

	// Open the background image
	Bitmap ^bkgrnd = nullptr;
	try {
		bkgrnd = gcnew Bitmap(args[0]);
	} catch (Exception ^ex) {
		Console::Error->WriteLine(L"There was a problem opening the background image '"+args[0]+L"': "+ex->Message);
		return 1;
	}

	// Target height and width
	int width = bkgrnd->Width, height = bkgrnd->Height;

	// Open the directory
	DirectoryInfo ^dir = nullptr;
	try {
		dir = gcnew DirectoryInfo(args[1]);
	} catch (Exception ^ex) {
		Console::Error->WriteLine(L"There was a problem opening the directory '"+args[1]+L"': "+ex->Message);
		return 1;
	}

	// Create/open destination directory
	String ^dest = args[2];
	if (File::Exists(dest)) {
		if (!Directory::Exists(dest)) {
			Console::Error->WriteLine(L"The path '"+dest+L"' is not a directory");
			return 1;
		}
	} else {
		try {
			Directory::CreateDirectory(dest);
		} catch (Exception ^ex) {
			Console::Error->WriteLine(L"There was a problem creating the directory '"+dest+L"': "+ex->Message);
			return 1;
		}
	}

	// Go through each file
	for each (FileInfo ^fi in dir->GetFiles(L"*.bmp")) {
		String ^name = fi->FullName;
		String ^file = Path::GetFileName(name);
		try {
			// Open it
			Bitmap ^img = gcnew Bitmap(name);
			if (img->Width != width || img->Height != height) {
				Console::WriteLine(L"Skipping '"+name+L"'.");
				continue;
			}
			// Go through each pixel
			for (int x = 0; x < width; x++) {
				for (int y = 0; y < height; y++) {
					// Get the foreground and background colors
					RGBA f = color2rgba(img->GetPixel(x, y));
					RGBA b = color2rgba(bkgrnd->GetPixel(x, y));
					// Convert to alpha, mix, and replace pixel
					img->SetPixel(x, y, rgba2color(mixColors(color2alpha(f, black), b)));
				}
			}
			img->Save(dest+L"\\"+file, Imaging::ImageFormat::Bmp);
		} catch (Exception ^ex) {
			Console::Error->WriteLine(L"There was a problem editing the image '"+file+L"': "+ex->Message);
		}
	}
	
    return 0;
}
